/* Based on:
 * simple driver for serial mouse
 * Andrew Haylett, 14th December 1992
 * and on the driver in XFree86.
 * Edited for svgalib (hhanemaa@cs.ruu.nl).
 * This probably doesn't work with all types of bus mouse.
 * HH: Added PS/2 mouse support.
 * Fixed Logitech support thanks to Daniel Jackson.
 * MouseSystems movement overflow fixed by Steve VanDevender.
 * Logitech fixed again.
 * Michael: Added support for controlling DTR and RTS.
 */

/* This file is included by mous_int.c. */

#ifndef _MS_C_
#define _MS_C_

#include <stdio.h>
#include <dpmi.h>
#include <sys/segments.h>

#include <time.h>

static int mouse_installed = 0;

typedef struct
{
	short int dx;
	short int dy;
	int b;
} Event;

#define EVENTS	1000

static Event event[EVENTS];
static int in = 0;
static int out = 0;
static short int oldx = 0;
static short int oldy = 0;

static _go32_dpmi_seginfo mouse_seginfo;
static _go32_dpmi_registers mouse_regs;

static void MouseInt(_go32_dpmi_registers *r)
{
	/* interrupt handler for int 33 */

	if ((in+1)%EVENTS == out)
	{
		/* overflow error */
	}
	else
	{
		event[in].b = ((r->x.bx >> 1) & 3) + ((r->x.bx & 1) << 2);
		event[in].dx = r->x.si - oldx; oldx = r->x.si;
		event[in].dy = r->x.di - oldy; oldy = r->x.di;
		in = (in+1)%EVENTS;
	}
}

END_OF_FUNCTION(MouseInt);

static inline void
ms_close() {
	__dpmi_regs r;

	if (!mouse_installed)
		return;

	r.x.ax = 0x0C;
	r.x.cx = 0;
	r.x.dx = 0;
	r.x.es = 0;
	__dpmi_int(MOUSE_INT, &r);         /* install NULL callback */

	r.x.ax = 15;
	r.x.cx = 8;
	r.x.dx = 16;
	__dpmi_int(MOUSE_INT, &r);         /* set sensitivity */

	/* remove callback */
	_go32_dpmi_free_real_mode_callback(&mouse_seginfo);
#ifdef DEBUG
	printf("mouse is closed\n");
#endif
	mouse_installed = 0;
}


static int
ms_init()
{
	/* initialize mouse via int MOUSE_INT (stolen from Allegro) */
	__dpmi_regs r;

	if (mouse_installed)
		return -1;

	r.x.ax = 0;
	__dpmi_int(MOUSE_INT, &r);         /* initialise mouse driver */

	if (r.x.ax != 0xffff)
		return -1;
		
	/* lock variables and function 'MouseInt' used as/in the handler */
	LOCK_VARIABLE(event);
	LOCK_VARIABLE(in);
	LOCK_VARIABLE(out);
	LOCK_VARIABLE(oldx);
	LOCK_VARIABLE(oldy);
	LOCK_VARIABLE(mouse_seginfo);
	LOCK_VARIABLE(mouse_regs);
	LOCK_FUNCTION(MouseInt);

	/* create real mode callback for the int33 driver (stolen from Allegro) */
	mouse_seginfo.pm_offset = (int)MouseInt;
	mouse_seginfo.pm_selector = _go32_my_cs();
	_go32_dpmi_allocate_real_mode_callback_retf(&mouse_seginfo, &mouse_regs);
	r.x.ax = 0x0C;
	r.x.cx = 0x7f;                      /* 3-button mice now supported! */
	r.x.dx = mouse_seginfo.rm_offset;
	r.x.es = mouse_seginfo.rm_segment;

	__dpmi_int(MOUSE_INT, &r);               /* install callback */

	mouse_installed = 1;

	return 0;
}



#define MOUSEBUFFERSIZE 256

static int
get_ms_event( int wait )
/*
    Changed to process multiple packets.
    wait value:
	0       Process any mouse events, and return status.
	1       Wait for mouse event, then return.

	Status indicates whether an event was processed.
*/
{
	/* this must be done because fill can be changed while running 
	this function */
/* ingo >>> */
	if (out != in)
	{
		while (out != in)
		{
		    	__MouseEventHandler(event[out].b,
		    				event[out].dx,event[out].dy);
			out = (out+1)%EVENTS;
		}

		return 1;
	}
	else
	{
		return 0;
	}
}

#endif
