/*
 * ASCD : copyright (c) Aley Keprt 1998-2001
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File fdi.c - VL 1772 disk controller emulation for SimCoupe
 *              contains also all disk support stuff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
  Big thanks to Simon Owen for fixing the problem with "disk-side-chaos".
  So, one of the very oldest bugs of SimCoupe/ASCD is gone (ASCD 0.95).
  It needed quite lots of additional changes, but it seems to work now.
  (Aley 11/2001)
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
typedef long blkcnt_t;
#include <sys/stat.h>
#include <unistd.h>
#include <zlib.h>

#include "fdi.h"

#ifndef SAMEMU
#include "util.h"
#include "resource.h"
#include "mem.h"

#include "sna.h"
#include "z80_snap.h"
#include "tape.h"
#include "fux.h"
#endif

#ifdef DJGPP
 #include "floppy.h"
#endif


//---------------------------------------------------------------------------


#define MAX_NO_SIDES	2
#define MAX_SECTOR_SIZE	6250   //6250=whole track


/* Default DOS system image */
#define SYSTEM_DOS "system.dos"


/* Disk image types */
enum {
  SAM_IMAGE,
  MSDOS_IMAGE,
  SAD_IMAGE,
  REAL_FD,
};


/* The various disk format image sizes */
#define SAM_FORMAT_SIZE   819200
#define MSDOS_FORMAT_SIZE 737280


/* Special path names for real floppy drives */
#define FDI_FLOPPY_1 "fd1:"

#define FDI_FLOPPY_2 "fd2:"


/* The ID string for Aley Keprt's SAD disk image format */
#define SAD_FORMAT_ID "Aley's disk backup"
#define SAD_ID_LENGTH (sizeof(SAD_FORMAT_ID)-1)



/* The VL1772 commands */
#define RESTORE		    0x00
#define SEEK		      0x10
#define STEP_NUPD	    0x20
#define STEP_UPD	    0x30
#define STEP_IN_NUPD	0x40
#define STEP_IN_UPD	  0x50
#define STEP_OUT_NUPD	0x60
#define STEP_OUT_UPD	0x70
#define READ_1SECTOR	0x80
#define READ_MSECTOR	0x90
#define WRITE_1SECTOR	0xa0
#define WRITE_MSECTOR	0xb0
#define READ_ADDRESS	0xc0
#define FORCE_INTERRUPT	0xd0
#define READ_TRACK	  0xe0
#define WRITE_TRACK	  0xf0


/* The possible status bits */
#define BUSY	    		0x01
#define DRQ		    	    0x02
#define LOST_DATA		    0x04
#define CRC_ERROR	    	0x08
#define RECORD_NOT_FOUND	0x10
#define SPIN_UP			    0x20
#define WRITE_PROTECT		0x40
#define MOTOR_ON	    	0x80


//---------------------------------------------------------------------------


extern uns8 keyports[];


typedef struct _vl1772_reg {
 uns8 Command;
 uns8 Status;
 uns8 Track;
 uns8 Sector;
 uns8 Data;
} vl1772_reg;


typedef struct _DiskAttr {
 int ImageType;
 uns8 sid;
 uns8 trk;
 uns8 sec;
 uns8 SectorDiv64;
} DiskAttr;


typedef struct {
 char  *FileName;
 FILE  *FPtr;

 uns8 *data;    //contents of a disk
 int changed;   //1=contents changed

 DiskAttr Dattr;
 vl1772_reg reg;
 uns8  SectorBuf[MAX_SECTOR_SIZE];
 uns8  *Bufp;
 int  ByteCount;
} Disk;


#define D_DISK(x) (((x) & 0x10) >> 4)
#define D_SIDE(x) (((x) & 0x04) >> 2)
#define DISK(x)  TheDisks[D_DISK(x)]


enum {COMMAND = 0, STATUS = 0, TRACK, SECTOR, DATA};


static Disk TheDisks[2];


int ReadSector (uns8 Side_no, Disk *disk);
int WriteSector(uns8 Side_no, Disk *disk);


//---------------------------------------------------------------------------


void fdi_out(int port, uns8 byte) {
 int Sdir = 1;

 switch(port&0x03) {
  
 case COMMAND:

  DISK(port).reg.Command=byte&0xf0;

  switch(byte&0xf0) {

  /* Type I commands */
  case RESTORE:
   DISK(port).reg.Track=0;
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP | LOST_DATA;
   break;

  case SEEK:
   DISK(port).reg.Track = DISK(port).reg.Data;
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
   if(!DISK(port).reg.Track) DISK(port).reg.Status |= LOST_DATA;
   break;

  case STEP_NUPD:
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
   break;

  case STEP_UPD:
   DISK(port).reg.Track += Sdir;
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
   if (!DISK(port).reg.Track) DISK(port).reg.Status |= LOST_DATA;
   break;

  case STEP_IN_NUPD:
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
   break;

  case STEP_IN_UPD:
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
   DISK(port).reg.Track++;
   break;

  case STEP_OUT_NUPD:
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
   break;

  case STEP_OUT_UPD:
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
   if (!DISK(port).reg.Track--) 
    DISK(port).reg.Status |= LOST_DATA;
   break;

  /* Type II Commands */
  case READ_1SECTOR:
  case READ_MSECTOR:
   DISK(port).reg.Status = BUSY | DRQ;
   if(ReadSector(D_SIDE(port),&DISK(port))) {
    DISK(port).reg.Status = RECORD_NOT_FOUND;
   }
   break;

  case WRITE_1SECTOR:
  case WRITE_MSECTOR:
   DISK(port).reg.Status = BUSY | DRQ;
   DISK(port).Bufp = DISK(port).SectorBuf;
   DISK(port).ByteCount=0;
   break;

  /* Type III Commands */
  case READ_ADDRESS:
   simcoupe_mesg(SIMCOUPE_WARN,"READ_ADDRESS is not supported");
   break;

  case READ_TRACK:
   simcoupe_mesg(SIMCOUPE_WARN,"READ_TRACK is not supported");
   break;

  case WRITE_TRACK:
   {
    static int info;
    if(!info++) simcoupe_mesg(SIMCOUPE_WARN,"WRITE_TRACK (i.e. disk formatting) is not supported");
   }
   break;

  /* Type IV Commands */
  case FORCE_INTERRUPT:
   DISK(port).reg.Status&=~BUSY;
   DISK(port).reg.Command=0;
   break;
   
  default:
   printf("Unknown Disk Command %x\n",byte & 0xf0);
   break;
  } /* End switch */
  break;
  
 case TRACK:
  DISK(port).reg.Track = byte;
  break;

 case SECTOR:
  DISK(port).reg.Sector = byte;
  break;

 case DATA:
  if(DISK(port).reg.Status & BUSY) {

   if(DISK(port).reg.Command == WRITE_1SECTOR ||
    DISK(port).reg.Command == WRITE_MSECTOR) {

    //write sector(s)

    *DISK(port).Bufp++ = byte;
/*    printf("ByteCount = %d\n",DISK(port).ByteCount);*/

    if(++DISK(port).ByteCount == (64*DISK(port).Dattr.SectorDiv64) ) {

     //sector is finished

     //flush buffer
     if(WriteSector(D_SIDE(port),&DISK(port))) {
      DISK(port).reg.Status = RECORD_NOT_FOUND;
      break;
     }

     //??? @@@ next line originally was READ_MSECTOR ???
     if(DISK(port).reg.Command == WRITE_MSECTOR &&
      DISK(port).reg.Sector++ != DISK(port).Dattr.sec) {
      DISK(port).Bufp = DISK(port).SectorBuf;
      DISK(port).ByteCount = 0;
     } else {
      DISK(port).reg.Status &= ~(BUSY | DRQ);
      DISK(port).reg.Command = 0;
     }
    }
   }
  }
  DISK(port).reg.Data=byte;
  break;
 }
}


//---------------------------------------------------------------------------


uns8 fdi_in(int port) {

 //Aley: This variable was uninitialized, so I expect "any" value is correct here.
 //Please correct me!
 uns8 byte=0xff;
 
 static uns8 DRQ_count=7;
 
 switch(port&0x03) {
  
 case STATUS :
/*  printf("Status : %x\n", DISK(port).reg.Status); */
  byte = DISK(port).reg.Status;
  if((!DISK(port).reg.Command) && !DRQ_count--) {
   DISK(port).reg.Status ^= DRQ;
   DRQ_count=7;
  }
  break;

 case TRACK:
  byte=DISK(port).reg.Track;
  break;

 case SECTOR:
  byte=DISK(port).reg.Sector;
  break;

 case DATA:
/*  printf("Data : %x\n",DISK(port).ByteCount); */
  if(DISK(port).reg.Status & BUSY) {
   if(DISK(port).reg.Command == READ_1SECTOR ||
    DISK(port).reg.Command == READ_MSECTOR) {

    //read sector(s)
    DISK(port).reg.Data=*DISK(port).Bufp++;
    if(++DISK(port).ByteCount == (64*DISK(port).Dattr.SectorDiv64)) {

     //sector is finished - do we have to read another one?
     if(DISK(port).reg.Command == READ_MSECTOR &&
      DISK(port).reg.Sector++ != DISK(port).Dattr.sec) {
      //move to next sector
      if(ReadSector(D_SIDE(port),&DISK(port))) {
       DISK(port).reg.Status = RECORD_NOT_FOUND;
       break;
      }
     } else {
      DISK(port).reg.Status &= ~(BUSY | DRQ);
      DISK(port).reg.Command = 0;
     }     
    }
   } 
  } 
  byte = DISK(port).reg.Data;
  break;
 } /* End Switch */

 return byte;
}


//---------------------------------------------------------------------------


uns8* load_disk_image(char *fn,DiskAttr *attr,int fsize) {
/*
  Check whether the disk image file fn is a SAD image
  Fill in the disk attributes from the file

  Loads disk into memory buffer, returns pointer to data
*/
 gzFile *gzf;
 char id[SAD_ID_LENGTH];
 long dsize;
 uns8 *data=0;

 gzf=gzopen(fn,"rb");
 if(!gzf) return NULL;

 if(gzread(gzf,id,SAD_ID_LENGTH)==SAD_ID_LENGTH) {

  if(!memcmp(id,SAD_FORMAT_ID,SAD_ID_LENGTH)) {
   attr->ImageType=SAD_IMAGE;
   attr->sid=gzgetc(gzf);
   attr->trk=gzgetc(gzf);
   attr->sec=gzgetc(gzf);
   attr->SectorDiv64=gzgetc(gzf);

   //disk image metrics must be nonzero
   if(!attr->sid || !attr->trk || !attr->sec || attr->SectorDiv64<128/64 || attr->SectorDiv64>MAX_SECTOR_SIZE/64) {
     //invalid disk metrics
     gzclose(gzf);
     return 0;
   }

   /* load data */
   dsize=attr->sid*attr->trk*attr->sec*attr->SectorDiv64*64;
   data=malloc(dsize);
   if(data) {
    if(gzread(gzf,data,dsize)==dsize) {
     gzclose(gzf);
#ifndef SAMEMU
     set_emulation_mode(EM_Sam);
#endif
     return data;
    }
   }
  } else {

   //SAD header not found - try other formats
   gzrewind(gzf);
   
   switch(fsize) {

   case SAM_FORMAT_SIZE:
     attr->ImageType = SAM_IMAGE;
     attr->sid = 2;
     attr->trk = 80;
     attr->sec = 10;
     attr->SectorDiv64 = 8;
     set_emulation_mode(EM_Sam);
     break;
     
   case MSDOS_FORMAT_SIZE:
     attr->ImageType = MSDOS_IMAGE;
     attr->sid = 2;
     attr->trk = 80;
     attr->sec = 9;
     attr->SectorDiv64 = 8;
     set_emulation_mode(EM_Sam);
     break;
   }
  }
 }

 if(data) free(data);
 gzclose(gzf);
 return 0;
}


//---------------------------------------------------------------------------


int isext(const char *fname,const char *ext) {
//
//return: 1=fname has extension ext
//
  return 0==stricmp(fname+strlen(fname)-strlen(ext),ext);
}


//---------------------------------------------------------------------------


void insert_disk(char *fname,int disk_no) {
//
//insert disk or load snapshot
//
 struct stat stbuff;
 DiskAttr tmpAttr;
 FILE *TheFPtr=0;
 uns8 *data=0;
 static int first=1;

#ifndef SAMEMU
 //snapshots are loaded elsewhere - detect them by file extension
 if(isext(fname,".sna") || isext(fname,".sna.gz")) {load_sna(fname);return;}
 if(isext(fname,".z80") || isext(fname,".z80.gz")) {load_z80(fname);return;}
 if(isext(fname,".fux") || isext(fname,".fux.gz")) {load_fux(fname);return;}
 if(isext(fname,".tap") || isext(fname,".tap.gz")) {open_tap(fname);return;}
#endif

 //the first inserted disk image is autobooted
 if(first) {
   first=0;
   keyports[2]&=0x7f;
 }

 //--------------------------------------------------------------------------

 disk_no--;

 /* Try to work out the type of disk image inserted */
 if(!stat(fname,&stbuff)) {
  
  if((stbuff.st_mode&S_IFMT)==S_IFREG) {
   //read header
   data=load_disk_image(fname,&tmpAttr,stbuff.st_size);
  }

  if(tmpAttr.sec>18) return;

  /* Pick up the File Pointer of the image */
  if(tmpAttr.ImageType!=SAD_IMAGE) TheFPtr=fopen(fname,"rb+");

 } else if(!strcmp(fname,FDI_FLOPPY_1) ) {

  /* Use the PC floppy drive 1 */ 
  tmpAttr.ImageType = REAL_FD;
  tmpAttr.sid = 0;  /* sid stores PC drive No */
  tmpAttr.trk = 80;
  tmpAttr.sec = 10;
  tmpAttr.SectorDiv64=512/64;
  
  TheFPtr = NULL;
#ifdef DJGPP
  FloppyInit(0);
#endif

 } else if(!strcmp(fname,FDI_FLOPPY_2) ) {

  /* Use the PC floppy drive 2 */ 
  tmpAttr.ImageType = REAL_FD;
  tmpAttr.sid = 1;  /* sid stores PC drive No */
  tmpAttr.trk = 80;
  tmpAttr.sec = 10;
  tmpAttr.SectorDiv64=512/64;
  
  TheFPtr = NULL;
#ifdef DJGPP
  FloppyInit(1);
#endif
 }

 //clear & init disk info
 memset(&TheDisks[disk_no],0,sizeof(TheDisks[disk_no]));
 TheDisks[disk_no].FileName=strdup(fname);
 TheDisks[disk_no].FPtr = TheFPtr;
 TheDisks[disk_no].Dattr = tmpAttr;
 TheDisks[disk_no].data=data;
 TheDisks[disk_no].changed=0;
}


void eject_disk(int disk_no) {

 Disk *side=&TheDisks[--disk_no];

 //perform delayed save, if necessary
 if(side->Dattr.ImageType==SAD_IMAGE && side->changed) {
  gzFile *gzf;
  struct _DiskAttr *attr=&side->Dattr;
  long dsize=attr->sid*attr->trk*attr->sec*attr->SectorDiv64*64;

  gzf=gzopen(side->FileName,"wb");
  if(gzf) {
   gzwrite(gzf,SAD_FORMAT_ID,SAD_ID_LENGTH);
   gzputc(gzf,attr->sid);
   gzputc(gzf,attr->trk);
   gzputc(gzf,attr->sec);
   gzputc(gzf,attr->SectorDiv64);

   /* save data */
   gzwrite(gzf,side->data,dsize);
   gzclose(gzf);
  }
 }

 free(side->FileName);
 side->FileName=0;
 
 if(side->FPtr) fclose(side->FPtr);
 side->FPtr=0;

 free(side->data);
 side->data=0;
}


int ReadSector(uns8 Side_no, Disk *Dside) {
 long offset;

 //vypne F9 (autoboot)
 keyports[2]|=0x80;

 if (Dside->FPtr || Dside->FileName) {

  switch(Dside->Dattr.ImageType) {

  case REAL_FD:
#ifdef __DJGPP
   FloppyRead(Dside->Dattr.sid,
              Side_no,
              Dside->reg.Track,
              Dside->reg.Sector,
              Dside->SectorBuf);
   Dside -> ByteCount = 0;
   Dside -> Bufp = Dside -> SectorBuf;
#endif
   return 0;
   
  case SAD_IMAGE:

   //check image metrics!
   if(Side_no>=Dside->Dattr.sid || Dside->reg.Track>=Dside->Dattr.trk || Dside->reg.Sector>Dside->Dattr.sec) {
     return -1;
   }

   offset =// SAD_HEADER +
    (64*Dside->Dattr.SectorDiv64) * 
    (Dside->Dattr.sec * (Dside->Dattr.trk*Side_no + Dside->reg.Track) +
    (Dside->reg.Sector - 1)
    );

   //move data from memory buffer
   memcpy(Dside->SectorBuf,Dside->data+offset,64*Dside->Dattr.SectorDiv64);

   //reset sector counters
   Dside -> ByteCount = 0;
   Dside -> Bufp = Dside -> SectorBuf;
   return 0;

  case SAM_IMAGE:
  case MSDOS_IMAGE:
   offset = ((64*Dside->Dattr.SectorDiv64) * Dside->Dattr.sec) * 
    (2*Dside->reg.Track + Side_no) +
    (64*Dside->Dattr.SectorDiv64)  * (Dside->reg.Sector - 1);
   break;

  default:
   return -1;
  }
  
  /*         printf("Side = %d,  Track = %d, Sector = %d, Offset = %d\n", 
  Side_no, Dside->reg.Track, Dside->reg.Sector, offset);
  */

  if (fseek(Dside->FPtr, offset, SEEK_SET) != 0) {
   return -2; /* couldn't seek the required sector */  
  } else {
   if(fread(Dside->SectorBuf, sizeof(uns8), (64*Dside->Dattr.SectorDiv64) , Dside->FPtr) != (64*Dside->Dattr.SectorDiv64) ) {
    return -3; /* couldn't read the required sector */
   } else {
   /*
   for (i = 0; i < 512 ; i++) 
   printf("%x",Dside->SectorBuf[i]);
   printf ("\n");
    */
    Dside -> ByteCount = 0;
    Dside -> Bufp = Dside -> SectorBuf;
    return 0; 
   }
  }
 } else {
  return -1; /* couldn't open image => drive not ready */
 }
}


int WriteSector(uns8 Side_no, Disk *Dside) {
  long offset;

  if(Dside->FPtr || Dside->FileName) {

  if(Dside->Dattr.ImageType == REAL_FD) {
#ifdef __DJGPP
   
   FloppyWrite(Dside->Dattr.sid,
               Side_no, Dside->reg.Track,
               Dside->reg.Sector,
               Dside->SectorBuf);
   
   Dside -> ByteCount = 0;
   Dside -> Bufp = Dside -> SectorBuf;
   return 0;
#endif
  }

  if(Dside->Dattr.ImageType == SAD_IMAGE) {

   //check image metrics!
   if(Side_no>=Dside->Dattr.sid || Dside->reg.Track>=Dside->Dattr.trk || Dside->reg.Sector>Dside->Dattr.sec) {
     return -1;
   }

   offset =// SAD_HEADER +
    (64*Dside->Dattr.SectorDiv64) * 
    (Dside->Dattr.sec * (Dside->Dattr.trk*Side_no + Dside->reg.Track) +
    (Dside->reg.Sector - 1)
    );

   //move data to the memory buffer
   memcpy(Dside->data+offset,Dside->SectorBuf,64*Dside->Dattr.SectorDiv64);

   //request delayed save
   Dside->changed=1;

   Dside->ByteCount=0;
   Dside->Bufp=Dside->SectorBuf;
   return 0;

  } else {

   offset = ((64*Dside->Dattr.SectorDiv64) * Dside->Dattr.sec) * 
    (2*Dside->reg.Track + Side_no) +
    (64*Dside->Dattr.SectorDiv64)  * (Dside->reg.Sector - 1);
   
  }
  /*         printf("Side = %d,  Track = %d, Sector = %d, Offset = %d\n", 
  Side_no, Dside->reg.Track, Dside->reg.Sector, offset);*/
  
  if(fseek(Dside->FPtr, offset, SEEK_SET) != 0) {
   return -2; /* couldn't seek the required sector */  
  } else {
   if(fwrite(Dside->SectorBuf, sizeof(uns8), (64*Dside->Dattr.SectorDiv64) , Dside->FPtr) != (64*Dside->Dattr.SectorDiv64) ) {
    return -3; /* couldn't write that sector */
   } else {
    fflush(Dside->FPtr);
    Dside -> ByteCount = 0;
    Dside -> Bufp = Dside -> SectorBuf;
    return 0;
   }
  }
 } else {
  return -1;
 }
}


int create_disk_image(char *ImageFile, int AddSystem) {
 /* Create a blank disk image (.dsk format) and */
 /* add a DOS system (SYSTEM_DOS) if required.  */
 int   iTrack;
 uns8 *TrackBuffer;
 FILE *ImageFp, *SystemFp;
 int   Status = 0;
 
 TrackBuffer = (uns8 *)malloc(5120*sizeof(uns8));
 
 memset(TrackBuffer, 0, 5120*sizeof(uns8));
 
 ImageFp=fopen(ImageFile,"wb");
 if(ImageFp) {
  /* First create a blank image */
  for (iTrack=0; iTrack < 160; iTrack++) {
   Status += !(fwrite(TrackBuffer, 512, 10, ImageFp) == 10);
  }
  
  /* Add the system if required */
  if(AddSystem) {
   rewind(ImageFp);
   SystemFp=fopen(SYSTEM_DOS,"rb");
   if(SystemFp) {
    /* First copy the directory entry */
    Status+=!(fread (TrackBuffer,256,1,SystemFp)==1);
    Status+=!(fwrite(TrackBuffer,256,1,ImageFp) ==1);
    
    /* Seek to the first data sector */
    Status+=fseek(ImageFp, 0xa000, SEEK_SET);
    Status+=!(fread (TrackBuffer,512,10,SystemFp)==10);
    Status+=!(fwrite(TrackBuffer,512,10,ImageFp) ==10);
    
    Status+=fseek(ImageFp, 0xc800, SEEK_SET);
    Status+=!(fread (TrackBuffer,512,10,SystemFp)==10);
    Status+=!(fwrite(TrackBuffer,512,10,ImageFp) ==10);
    fclose(SystemFp);
   } else {
    Status = 1;    
   }
  }
  fclose(ImageFp);
 }
 else {
  Status = 1;
 }
 
 free(TrackBuffer);
 return Status;
}