/* 
 * ASCD : copyright (c) Aley Keprt 1998-2001
 *
 * File kbd.c : Keyboard emulation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "z80.h"
#include "events.h"
#include "util.h"
#include "kbd.h"
#include "keyb_int.h"
#include "samio.h"
#include "resource.h"
#include "joy.h"
#include "video.h"
#include "intrface.h"
#include "mouse.h"

#ifdef USE_AIR
#include "AirClient.h"
#endif


uns8 keyports[10];      //8 rows of keys (+ 1 row for unscanned keys)

int SpectrumKeyboard;   //switch for spectrum keyboard mode

static int shiftkey;//flag for shift key
static int ctrlkey; //flag used for handling Ctrl+F1,F2,F3,etc.

static int EmulatorControlEvent;


#ifdef USE_AIR
  #define recording() (Air_getmode()==AM_record)
#else
  #define recording() 0
#endif


//----------------------------------------------------------------------------


/* Maps vgalib scancodes to a Sam keyboard row and bit to set/reset */
/* E.g. ktbl[SCAN_A] means set/reset bit 0 (0xfe) of keyboard row 1 */
static struct ktbl {
	int port;       /* which keyboard row is the key on? */
	int bit;        /* which bit to mask */
} keytable[51] = {
	{9, 0},

  /* Esc */
  {9, 0},

	/* 1 2 3 4 5 6 7 8 9 0 */
	{3, 0xfe}, {3, 0xfd}, {3, 0xfb}, {3, 0xf7}, {3, 0xef}, 
	{4, 0xef}, {4, 0xf7}, {4, 0xfb}, {4, 0xfd}, {4, 0xfe},

  /* - + Backspace Tab */
	{9, 0}, {9, 0}, {9, 0}, {9,0},

	/* Q W E R T Y U I O P */
	{2, 0xfe}, {2, 0xfd}, {2, 0xfb}, {2, 0xf7}, {2, 0xef},
	{5, 0xef}, {5, 0xf7}, {5, 0xfb}, {5, 0xfd}, {5, 0xfe},

	/* [ ] Enter Ctrl */
	{9, 0}, {9, 0}, {9, 0}, {9,0},

	/* A S D F G H J K L */
	{1, 0xfe}, {1, 0xfd}, {1, 0xfb}, {1, 0xf7}, {1, 0xef},
	{6, 0xef}, {6, 0xf7}, {6, 0xfb}, {6, 0xfd},
	
	{9, 0}, {9, 0}, {9, 0}, {9,0}, {9,0},

	/* Z X C V B N M */
	{0, 0xfd}, {0, 0xfb}, {0, 0xf7}, {0, 0xef},
	{7, 0xef}, {7, 0xf7}, {7, 0xfb}
};


//#define FK_HELP SCAN_F1
#define FK_MENU SCAN_F2
//#define FK_DISK1 SCAN_F3
//#define FK_DISK2 SCAN_F4

#define FK_VIDEOMODE SCAN_F5
#define FK_MOUSEMODE SCAN_F6
#define FK_KEYMODE SCAN_F7

#define FK_SYNC SCAN_F9
#define FK_SCREENSHOT SCAN_F10
#define FK_NMI SCAN_F11
#define FK_RESET SCAN_F12


//----------------------------------------------------------------------------


#define and(port,bit) (keyports[port]&=~bit)
#define pressed 1
#define notshift if(!shiftkey)
#define ifelse else
#define relshift (keyports[0]|=1)
static void sam_keypress(int ks)
#include "samkbd.h"
static void zx_keypress(int ks)
#include "zxkbd.h"


#define and(port,bit) (keyports[port]|=bit)
#define pressed 0
#define notshift
#define ifelse
//relshift zde ma zmacknout shift pri uvolneni jine klavesy, lae jen
//v pripade, ze shiftkey ukazuje na to, ze shift je skutecne zmacknuty
#define relshift (shiftkey ? keyports[0]&=~1 : 0)
static void sam_keyrelease(int ks)
#include "samkbd.h"
static void zx_keyrelease(int ks)
#include "zxkbd.h"


void KbdUpdate(void) {
//
//volano pri preruseni
//zpracuje klavesy v bufferu klavesnice
//
  ProcessKeyBuf();
}


/* Emulator control keys - dealt with once per frame, and flagged */
/* by interrupt driven keyboard handler	                          */
void EmulatorControl(void) {

	int i;
	
	if(EmulatorControlEvent) {
    switch(EmulatorControlEvent) {

    //F2 = menu
		case FK_MENU:
      if(!recording()) ui_enter();
			break;
/*
		case FK_DISK1:
			if(!UiActive) ui_main(UI_DISK1);
			break;

		case FK_DISK2:
			if(!UiActive) ui_main(UI_DISK2);
			break;
*/
    //F5 = display mode
		case FK_VIDEOMODE:
      //Ctrl+F5 switches autoscale
			if(ctrlkey) {
        osd(VideoSwitchAutoScale() ? "AutoScale: On" : "AutoScale: Off");
      } else {
        VideoSwitchScale();
        /* Call vmpr to reset the screen write pointers
           and fetch the screen from memory */
        out_vmpr(vmpr);
      }
			break;

    //F6 = mouse mode
		case FK_MOUSEMODE:
			if(ctrlkey) {
  			//switch mouse resolution
				if(SAMPointerMode) {
          SAMPointerMode=3-SAMPointerMode;
          osd(SAMPointerMode==2 ? "Mouse: HiRes" : "Mouse: LoRes");
        }
			} else {
  			//mouse on/off
			  if(!SAMPointerMode) {
				  SAMPointerMode=1;
  				sam_mouse_button_write(PointerButtonsOff);
	  			sam_mouse_xy_write(0, 0);
          osd("Mouse: On");
			  } else {
				  SAMPointerMode=0;
          osd("Mouse: Off");
	  		}
      }
      break;

    //F7 = keyboard mode
		case FK_KEYMODE:
			SpectrumKeyboard^=1;
      osd(SpectrumKeyboard ? "Keyboard: ZX Spectrum" : "Keyboard: Sam Coupe");
			break;

    //F9 = 50Hz sync on/off
    case FK_SYNC:
      if(!recording()) {
        cfg.sync^=1;
        osd(cfg.sync ? "Sync: on" : "Sync: off");
      }
      break;

    //F10 = screenshot without border
    //Ctrl+F10 = screenshot with border
		case FK_SCREENSHOT:
			save_screenshot(ctrlkey);
      osd(ctrlkey ? "Screenshot with border" : "Screenshot");
			break;

    //F11 = NMI
		case FK_NMI:
      if(!recording()) {
        InterruptType=Z80_nmi;
        osd("NMI");
      }
			break;

 		//F12 = reset emulator
    //Ctrl+F12 = exit emulator
		case FK_RESET:
			if(ctrlkey) Quit(0);

      if(!recording()) {
        //reset sometimes didn't work, so I've added page_in combo here
        //and hoping it will solve the problems (Aley, ASCD 0.95)
        InterruptType=Z80_reset;
        for(i=0;i<9;i++) keyports[i]=0xff;
        out_byte(250,0);
        out_zxpage(0);

        osd("Reset");
      }
			break;
    }
	}
	EmulatorControlEvent = False;
}


void KbdInit(void) {
//
//keyboard initialization
//
  int i;

  if(KeyboardInit(SamKeyboardHandler)) {
    simcoupe_mesg(SIMCOUPE_FATAL,"Can't initialize keyboard");
  }
  OnQuit(KeyboardClose,"KeyboardClose()");

  //no keys initially being pressed
  for(i=0;i<9;i++) keyports[i]=0xff;
}


//-----------------------------------------------------------------------------


void SamKeyboardHandler(int scan, int EventType) {
//
//keyboard handler
//
//control is passed to here if a keyboard event is trapped
//by the interrupt handler
//
  switch(EventType) {
	
  case KEY_EVENTPRESS:
    if((scan>=SCAN_F1 && scan<=SCAN_F10) ||
       scan==SCAN_F11 || scan==SCAN_F12) {
      EmulatorControlEvent=scan;
    } else {
      if(!SpectrumKeyboard) sam_keypress(scan); else zx_keypress(scan);
    }
    break;

  case KEY_EVENTRELEASE:
    if(!SpectrumKeyboard) sam_keyrelease(scan); else zx_keyrelease(scan);
    break;
    
  default:
    simcoupe_mesg(SIMCOUPE_ERR,"Unknown Key event type");
  }
}

