/* 
 * ASCD : copyright (c) Aley Keprt 1998-2001
 *
 * File resource.c : Configuration data
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <string.h>
#include <memory.h>

#include "keyb_int.h"
#include "mous_int.h"

#include "version.h"
#include "resource.h"
#include "util.h"
#include "speaker.h"
#include "mem.h"

#ifdef USE_AIR
#include "AirClient.h"
#endif


#define ASCD_INI "ascd.ini"


#define item(type,name,switch,def,help) def,
CFG cfg={
#include "cfg_vars.h"
};


#define item(type,name,switch,def,help) C_##name,
enum CFG_enum {
#include "cfg_vars.h"
  C_last,
};


typedef enum {ST_INT, ST_BOOL, ST_TEXT} SwitchType;
#define item(type,name,switch,def,help) {ST_##type,switch,help},
struct CfgText {
  SwitchType type;
  char *name;
  char *help;
} cfgpar[]={
#include "cfg_vars.h"
};


static void usage(void) {
  show_credits();

  list_switches();
  
  printf("\nFile types (can be packed with gzip): .sad, .dsk, .sna, .z80, .fux, .tap\n");
  
  exit(0);
}


//---------------------------------------------------------------------------


static void save_ini(void) {
//
//save configuration file
//
  FILE *f;
  int i;

  char fname[1000];
  sprintf(fname,"%s\\" ASCD_INI,exe_path);

  f=fopen(fname,"w");
  if(!f) simcoupe_mesg(SIMCOUPE_ERR,"Cannot write configuration file. Disk full or write protexted?");

  fprintf(f,";ASCD " VERSION_TEXT " configuration file - (c) by Aley Keprt\n"
            ";http://www.keprt.cz, aley@keprt.cz\n"
            ";type \"ascd -help\" to see complete list of options\n\n");

  for(i=4;i<C_last;i++) if(cfgpar[i].name) {

    switch(cfgpar[i].type) {

    case ST_BOOL:
      fprintf(f,"%s%-8s\t;%s%s%s (bool)\n",((long*)&cfg)[i] ? "" : "no",
              cfgpar[i].name,
              cfgpar[i].help[0]=='/' ? "-" : "",
              cfgpar[i].help[0]=='/' ? cfgpar[i].name : "",
              cfgpar[i].help);
      break;

    case ST_INT:
      fprintf(f,"%s %i\t;%s\n",
              cfgpar[i].name,((long*)&cfg)[i],cfgpar[i].help);
      break;

    case ST_TEXT:
      fprintf(f,"%s \"%s\"\t;%s\n",
              cfgpar[i].name,(char*)(((long*)&cfg)[i]),cfgpar[i].help);
      break;
    }
  }

  fclose(f);
}


//---------------------------------------------------------------------------


static int is_true(char *str) {
  return str ? !strnicmp(str,"true",4) || !strnicmp(str,"on",2) ||
               !strnicmp(str,"yes",3) || !strnicmp(str,"1",1) : 0;
}


//---------------------------------------------------------------------------


static int scan_int(char *str) {
  int i;
  if(!sscanf(str,"%d",&i)) {
    simcoupe_mesg(SIMCOUPE_FATAL,"Error: Requested a number: %s",str);
    Quit(1);
  }
  return(i);
}


//---------------------------------------------------------------------------


static void read_cmdline(int argc, char **argv) {
//
//read command line
//
  int i;
  int iarg=argc;
  argv++; //skip exe path

  while(--iarg) {
    if(argv[0][0]=='-' || argv[0][0]=='/') {

      int bval=1; //bool value
      const char *p=argv[0]+1; //p=switch name

      if(!strncmp("no",p,2)) {
        p+=2;
        bval=0;
      }

      //najde switch podle tabulky jmen
      for(i=0;i<C_last;i++) {
         if(cfgpar[i].name && !stricmp(cfgpar[i].name,p)) {
        switch(cfgpar[i].type) {

        case ST_BOOL:
          ((long*)&cfg)[i]=bval;
          argv++;
          break;

        case ST_INT:
          if(--iarg) {
            ((long*)&cfg)[i]=scan_int(argv[1]);
            argv+=2;
          }
          break;

        case ST_TEXT:
          if(--iarg) {
            strcpy((char*)(((long*)&cfg)[i]),argv[1]);
            argv+=2;
          }
          break;

        default:
          continue;
        }
        break;
      }
      }

      if(i>=C_last) {
        //custom switches
        if(!stricmp(p,"help")) {
          usage();
          exit(0);
        }
        else
#ifdef USE_AIR
        if(!stricmp(p,"record") && --iarg) {
          SetCfgStr(record,argv[1]);
          argv+=2;
        }
        else
        if(!stricmp(p,"playback") && --iarg) {
          SetCfgStr(playback,argv[1]);
          argv+=2;
        }
        else
        if(!stricmp(p,"scode") && --iarg) {
          cfg.Scode=atoi(argv[1]);
          argv+=2;
        }
        else
#endif
        if(!stricmp(p,"zx")) {
          cfg.zxmode=EM_48k;
          argv++;
        }
        else
        if(!stricmp(p,"128")) {
          cfg.zxmode=EM_128k;
          argv++;
        }
        else
        if(!stricmp(p,"sam")) {
          cfg.zxmode=EM_Sam;
          argv++;
        }
        else {
          //error
          printf("Invalid command line switch: %s\n",argv[0]);
          exit(0);
        }
      }

    } else {

      //by Aley Keprt 2000
      //tries to open named file
      FILE *f;

      //remove doublequotes
      if(argv[0][0]=='"' && argv[0][strlen(argv[0])-1]=='"') {
        argv[0]++;
        argv[0][strlen(argv[0])-1]=0;
      }

      f=fopen(*argv,"rb");
      if(f) {
        //found - use it as disk 1
        fclose(f);
        SetCfgStr(disk1,*argv);
        argv++;
      } else {
        //not found --> error
        printf("Error: Cannot open specified file \"%s\"",argv[0]);
        exit(0);
      }
    }
  }       
}


//---------------------------------------------------------------------------


char* clean(char *t) {
//
//remove trailing cr/lf if any
//remove leading and trailing spaces and tabs
//
  char *end;

  if(!t) return 0;

  //skip leading spaces
  while(t[0]==' ' || t[0]=='\t') t++;

  end=t+strlen(t)-1;

  while(end>=t && (end[0]=='\r' || end[0]=='\n' || end[0]==' ' || end[0]=='\t')) {
    end[0]=0;
    end--;
  }

  return t;
}


//---------------------------------------------------------------------------


static void load_ini(void) {
//
//load ini file
//perse it using read_cmdline()
//
  char *argv[4];
  FILE *f;
  char text[1000];
  char *end;

  char fname[1000];
  sprintf(fname,"%s\\" ASCD_INI,exe_path);

  f=fopen(fname,"r");
  if(!f) return;

  while(1) {

    if(!fgets(text+1,sizeof(text)-1,f)) break; //error or eof

    //change first semicolon to \0
    end=strchr(text+1,';');
    if(end) *end=0;

    //argv[1]=switch
    argv[1]=text+1;

    //argv[2]=parameter
    argv[2]=strchr(argv[1],' ');
    if(argv[2]) {
      //split switch and parameter
      argv[2][0]=0;
      argv[2]++;

      //clean parameter
      argv[2]=clean(argv[2]);

      //remove doublequotes around parameter
      if(argv[2][0]=='"') argv[2]++;
      end=argv[2]+strlen(argv[2]);
      if(end[-1]=='"') end[-1]=0;
    }

    //clean switch
    argv[1]=clean(argv[1]);
    if(argv[1][0]!=0 && argv[1][0]!='-') {
      argv[1]--;
      argv[1][0]='-';
    }

    argv[0]="exe";
    argv[3]=0;
    if(argv[1][0]) read_cmdline(2+(argv[2] && argv[2][0]!=0),argv);
  }

  fclose(f);
}


//---------------------------------------------------------------------------


void init_cfg(int argc, char **argv) {
//
//initialize configuration - read ini file and command line
//
  int i;

  //alokuje pamet pro texty
  for(i=0;i<C_last;i++) if(cfgpar[i].type==ST_TEXT) {
    char *p=(char*)malloc(260);
    if(((long*)&cfg)[i]) strcpy(p,(char*)(((long*)&cfg)[i])); else p[0]=0;
    ((long*)&cfg)[i]=(long)p;
  }

  load_ini();
  read_cmdline(argc,argv);

#ifdef USE_AIR
  if(cfg.record[0]) cfg.playback[0]=0; //nemuze byt zaroven record i playback
  if(cfg.record[0] || cfg.playback[0]) setairoptions();
#endif

  OnQuit(save_ini,"save_ini()");
}


//---------------------------------------------------------------------------


void list_switches(void) {
  int i;

  printf("-help - display this help\n"
         "-record <fname> - record input to named file\n"
         "-playback <fname> - playback recorded input from named file\n"
         "-sam - Sam Coupe emulation\n"
         "-zx - ZX Spectrum emulation\n"
         "-128 - ZX Spectrum 128 emulation\n"
         );

  for(i=0;i<C_last;i++) if(cfgpar[i].name && cfgpar[i].help) {
    printf("-%s%s\n",cfgpar[i].name,cfgpar[i].help);
  }
}

